/*++
/* NAME
/*	mynetworks 3
/* SUMMARY
/*	generate patterns for my own interface addresses
/* SYNOPSIS
/*	#include <mynetworks.h>
/*
/*	const char *mynetworks()
/* DESCRIPTION
/*	This routine uses the address list built by own_inet_addr()
/*	to produce a list of patterns that match the corresponding
/*	networks.
/*
/*	The interface list is specified with the "inet_interfaces"
/*	configuration parameter.
/*
/*	The address to netblock conversion style is specified with
/*	the "mynetworks_style" parameter: one of "class" (match
/*	whole class A, B, C or D networks), "subnet" (match local
/*	subnets), or "host" (match local interfaces only).
/* LICENSE
/* .ad
/* .fi
/*	The Secure Mailer license must be distributed with this software.
/* AUTHOR(S)
/*	Wietse Venema
/*	IBM T.J. Watson Research
/*	P.O. Box 704
/*	Yorktown Heights, NY 10598, USA
/*
/*	Dean C. Strik
/*	Department ICT Services
/*	Eindhoven University of Technology
/*	P.O. Box 513
/*	5600 MB  Eindhoven, Netherlands
/*	E-mail: <dean@ipnet6.org>
/*--*/

/* System library. */

#include <sys_defs.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#ifndef IN_CLASSD_NET
#define IN_CLASSD_NET		0xf0000000
#define IN_CLASSD_NSHIFT 	28
#endif

#define BITS_PER_ADDR_V4	32
#define BITS_PER_ADDR_V6	128

/* Utility library. */

#include <msg.h>
#include <vstring.h>
#include <inet_addr_list.h>
#include <name_mask.h>
#ifdef INET6
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#endif

/* Global library. */

#include <own_inet_addr.h>
#include <mail_params.h>
#include <mynetworks.h>

/* Application-specific. */

#define MASK_STYLE_CLASS	(1 << 0)
#define MASK_STYLE_SUBNET	(1 << 1)
#define MASK_STYLE_HOST		(1 << 2)

static NAME_MASK mask_styles[] = {
    MYNETWORKS_STYLE_CLASS, MASK_STYLE_CLASS,
    MYNETWORKS_STYLE_SUBNET, MASK_STYLE_SUBNET,
    MYNETWORKS_STYLE_HOST, MASK_STYLE_HOST,
    0,
};

/* mynetworks - return patterns that match my own networks */

const char *mynetworks(void)
{
    static VSTRING *result;
    int bits_per_addr;
#ifdef INET6
    char hbuf[NI_MAXHOST];
#endif

    if (result == 0) {
	char   *myname = "mynetworks";
	INET_ADDR_LIST *my_addr_list;
	INET_ADDR_LIST *my_mask_list;
	unsigned long addr = 0;
	unsigned long mask = 0;
	struct in_addr net;
	int     shift = 0;
	int     junk;
	int     i;
	int     mask_style;
#ifdef INET6
	struct sockaddr *sa;
#endif

	mask_style = name_mask("mynetworks mask style", mask_styles,
			       var_mynetworks_style);

	/*
	 * XXX Workaround: name_mask() needs a flags argument so that we can
	 * require exactly one value, or we need to provide an API that is
	 * dedicated for single-valued flags.
	 */
	for (i = 0, junk = mask_style; junk != 0; junk >>= 1)
	    i += (junk & 1);
	if (i != 1)
	    msg_fatal("bad %s value: %s; specify exactly one value",
		      VAR_MYNETWORKS_STYLE, var_mynetworks_style);

	result = vstring_alloc(20);
	my_addr_list = own_inet_addr_list();
	my_mask_list = own_inet_mask_list();

	for (i = 0; i < my_addr_list->used; i++) {
#ifdef INET6
	    sa = (struct sockaddr *)&my_addr_list->addrs[i];
	    if (sa->sa_family != AF_INET && sa->sa_family != AF_INET6) {
		msg_warn("%s: unknown family in address list", myname);
		 continue;
	    }
	    if (sa->sa_family == AF_INET) {
		bits_per_addr = BITS_PER_ADDR_V4;
		addr = ntohl(((struct sockaddr_in *)sa)->sin_addr.s_addr);
		mask = ntohl(((struct sockaddr_in *)
			      &my_mask_list->addrs[i])->sin_addr.s_addr);
	    } else
		bits_per_addr = BITS_PER_ADDR_V6;
#else
	    addr = ntohl(my_addr_list->addrs[i].s_addr);
	    mask = ntohl(my_mask_list->addrs[i].s_addr);
#endif

	    switch (mask_style) {

		/*
		 * Natural mask. This is dangerous if you're customer of an
		 * ISP who gave you a small portion of their network.
		 */
	    case MASK_STYLE_CLASS:
#ifdef INET6
		if (sa->sa_family == AF_INET) {
#endif
		if (IN_CLASSA(addr)) {
		    mask = IN_CLASSA_NET;
		    shift = IN_CLASSA_NSHIFT;
		} else if (IN_CLASSB(addr)) {
		    mask = IN_CLASSB_NET;
		    shift = IN_CLASSB_NSHIFT;
		} else if (IN_CLASSC(addr)) {
		    mask = IN_CLASSC_NET;
		    shift = IN_CLASSC_NSHIFT;
		} else if (IN_CLASSD(addr)) {
		    mask = IN_CLASSD_NET;
		    shift = IN_CLASSD_NSHIFT;
		} else {
#ifdef INET6
		    if (getnameinfo(sa, SA_LEN(sa), hbuf, sizeof(hbuf),
				    NULL, 0, NI_NUMERICHOST))
			strncpy(hbuf, "???", sizeof(hbuf));
		    msg_fatal("%s: bad address class: %s", myname, hbuf);
#else
		    msg_fatal("%s: bad address class: %s",
			      myname, inet_ntoa(my_addr_list->addrs[i]));
#endif
		}
		break;
#ifdef INET6
		} /* if AF_INET */
		/*
		 * There are no classes for IPv6, we default to subnets instead.
		 */
		/* FALLTHROUGH */
#endif

		/*
		 * Subnet mask. This is safe, but breaks backwards
		 * compatibility when used as default setting.
		 */
	    case MASK_STYLE_SUBNET:
#ifdef INET6
		if (sa->sa_family == AF_INET6) {
		    unsigned char *ac, *end;
		    ac = (unsigned char *)&(((struct sockaddr_in6 *)&my_mask_list->addrs[i])->sin6_addr);
		    end = ac + bits_per_addr / 8;
		    shift = bits_per_addr;
		    while (ac < end) {
			switch (*(ac++)) {
			    case 0xff: shift -= 8; break;
			    case 0xfe: shift -= 7; break;
			    case 0xfc: shift -= 6; break;
			    case 0xf8: shift -= 5; break;
			    case 0xf0: shift -= 4; break;
			    case 0xe0: shift -= 3; break;
			    case 0xc0: shift -= 2; break;
			    case 0x80: shift -= 1; break;
			    case 0x00: break;
			    default: msg_fatal("%s: inconsistent prefixlen",
				myname);
			}
		    }
		    break;
		}
#endif
		/* AF_INET */
		junk = mask;
		shift = bits_per_addr;
		while (junk != 0) {
		    shift--;
		    junk <<= 1;
		}
		break;

		/*
		 * Host only. Do not relay authorize other hosts.
		 */
	    case MASK_STYLE_HOST:
#ifdef INET6
		if (sa->sa_family == AF_INET6) {
		    shift = 0;
		    break;
		}
#endif
		mask = ~0;
		shift = 0;
		break;

	    default:
		msg_panic("unknown mynetworks mask style: %s",
			  var_mynetworks_style);
	    }
#ifdef INET6
	    if (sa->sa_family == AF_INET6) {
		if (getnameinfo(sa, SA_LEN(sa), hbuf, sizeof(hbuf), NULL, 0,
				NI_NUMERICHOST))
		    msg_fatal("%s: bad address to getnameinfo()", myname);
		vstring_sprintf_append(result, "[%s]/%d ",
				       hbuf, bits_per_addr - shift);
		continue;
	    }
#endif
	    /* AF_INET */
	    net.s_addr = htonl(addr & mask);
	    vstring_sprintf_append(result, "%s/%d ",
				   inet_ntoa(net), bits_per_addr - shift);
	}
	if (msg_verbose)
	    msg_info("%s: %s", myname, vstring_str(result));
    }
    return (vstring_str(result));
}

#ifdef TEST

char   *var_inet_interfaces;
char   *var_mynetworks_style;

int     main(int argc, char **argv)
{
    if (argc != 3)
	msg_fatal("usage: %s mask_style interface_list", argv[0]);
    msg_verbose = 10;
    var_inet_interfaces = argv[2];
    var_mynetworks_style = argv[1];
    mynetworks();
}

#endif
