/*++
/* NAME
/*	match_ops 3
/* SUMMARY
/*	simple string or host pattern matching
/* SYNOPSIS
/*	#include <match_ops.h>
/*
/*	int	match_string(flags, string, pattern)
/*	int	flags;
/*	const char *string;
/*	const char *pattern;
/*
/*	int	match_hostname(flags, name, pattern)
/*	int	flags;
/*	const char *name;
/*	const char *pattern;
/*
/*	int	match_hostaddr(flags, addr, pattern)
/*	int	flags;
/*	const char *addr;
/*	const char *pattern;
/* DESCRIPTION
/*	This module implements simple string and host name or address
/*	matching. The matching process is case insensitive. If a pattern
/*	has the form type:name, table lookup is used instead of string
/*	or address comparison.
/*
/*	match_string() matches the string against the pattern, requiring
/*	an exact (case-insensitive) match. The flags argument is not used.
/*
/*	match_hostname() matches the host name when the hostname matches
/*	the pattern exactly, or when the pattern matches a parent domain
/*	of the named host. The flags argument specifies the bit-wise OR
/*	of zero or more of the following:
/* .IP MATCH_FLAG_PARENT
/*	The hostname pattern foo.com matches itself and any name below
/*	the domain foo.com. If this flag is cleared, foo.com matches itself
/*	only, and .foo.com matches any name below the domain foo.com.
/* .RE
/*	Specify MATCH_FLAG_NONE to request none of the above.
/*
/*	match_hostaddr() matches a host address when the pattern is
/*	identical to the host address, or when the pattern is a net/mask
/*	that contains the address. The mask specifies the number of
/*	bits in the network part of the pattern. The flags argument is
/*	not used.
/* LICENSE
/* .ad
/* .fi
/*	The Secure Mailer license must be distributed with this software.
/* AUTHOR(S)
/*	Wietse Venema
/*	IBM T.J. Watson Research
/*	P.O. Box 704
/*	Yorktown Heights, NY 10598, USA
/*
/*	Takahiro Igarashi
/*
/*	Dean C. Strik
/*	Department ICT Services
/*	Eindhoven University of Technology
/*	P.O. Box 513
/*	5600 MB  Eindhoven, Netherlands
/*	E-mail: <dean@ipnet6.org>
/*--*/

/* System library. */

#include <sys_defs.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#ifdef INT_MAX_IN_LIMITS_H
#include <limits.h>
#endif

#ifdef STRCASECMP_IN_STRINGS_H
#include <strings.h>
#endif

#ifndef INADDR_NONE
#define INADDR_NONE 0xffffffff
#endif

/* Utility library. */

#include <msg.h>
#include <msg_output.h>
#include <mymalloc.h>
#include <split_at.h>
#include <dict.h>
#include <match_ops.h>
#include <stringops.h>

#define BITS_PER_ADDR_V4	32
#define BITS_PER_ADDR_V6	128

#ifdef INET6

/*
 * IPv6-enabled code was written by Takahiro Igarashi and Dean Strik.
 */

#endif /* INET6 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <resolv.h>

/* prototypes */
static PRINTFLIKE(2,3) void warning_msg(VSTRING *, const char *, ...);
#ifdef INET6
static int mask_comp(void *, void *, int);
#endif /* INET6 */

/* match_string - match a string literal */

int     match_string(int unused_flags, const char *string, const char *pattern)
{
    char   *myname = "match_string";
    int     match;
    char   *key;

    if (msg_verbose)
	msg_info("%s: %s ~? %s", myname, string, pattern);

    /*
     * Try dictionary lookup: exact match.
     */
    if (strchr(pattern, ':') != 0) {
	key = lowercase(mystrdup(string));
	match = (dict_lookup(pattern, key) != 0);
	myfree(key);
	if (match != 0)
	    return (1);
	if (dict_errno != 0)
	    msg_fatal("%s: table lookup problem", pattern);
	return (0);
    }

    /*
     * Try an exact string match.
     */
    if (strcasecmp(string, pattern) == 0) {
	return (1);
    }

    /*
     * No match found.
     */
    return (0);
}

/* match_hostname - match a host by name */

int     match_hostname(int flags, const char *name, const char *pattern)
{
    char   *myname = "match_hostname";
    const char *pd;
    const char *entry;
    char   *next;
    char   *temp;
    int     match;

    if (msg_verbose)
	msg_info("%s: %s ~? %s", myname, name, pattern);

    /*
     * Try dictionary lookup: exact match and parent domains.
     */
    if (strchr(pattern, ':') != 0) {
	temp = lowercase(mystrdup(name));
	match = 0;
	for (entry = temp; *entry != 0; entry = next) {
	    if ((match = (dict_lookup(pattern, entry) != 0)) != 0)
		break;
	    if (dict_errno != 0)
		msg_fatal("%s: table lookup problem", pattern);
	    if ((next = strchr(entry + 1, '.')) == 0)
		break;
	    if (flags & MATCH_FLAG_PARENT)
		next += 1;
	}
	myfree(temp);
	return (match);
    }

    /*
     * Try an exact match with the host name.
     */
    if (strcasecmp(name, pattern) == 0) {
	return (1);
    }

    /*
     * See if the pattern is a parent domain of the hostname.
     */
    else {
	if (flags & MATCH_FLAG_PARENT) {
	    pd = name + strlen(name) - strlen(pattern);
	    if (pd > name && pd[-1] == '.' && strcasecmp(pd, pattern) == 0)
		return (1);
	} else if (pattern[0] == '.') {
	    pd = name + strlen(name) - strlen(pattern);
	    if (pd > name && strcasecmp(pd, pattern) == 0)
		return (1);
	}
    }
    return (0);
}

#ifndef INET6
/* match_parse_mask - parse net/mask pattern */

static int match_parse_mask(const char *pattern, unsigned long *net_bits,
			            unsigned int *mask_shift)
{
    char   *saved_pattern;
    char   *mask;

    saved_pattern = mystrdup(pattern);
    if ((mask = split_at(saved_pattern, '/')) != 0) {
	if (!alldig(mask) || (*mask_shift = atoi(mask)) > BITS_PER_ADDR_V4
	    || (*net_bits = inet_addr(saved_pattern)) == INADDR_NONE) {
	    msg_fatal("bad net/mask pattern: %s", pattern);
	}
    }
    myfree(saved_pattern);
    return (mask != 0);
}

#endif

static void PRINTFLIKE(2,3) warning_msg(VSTRING *vp, const char *fmt,...)
{
    va_list ap;
    if (vp) {
	va_start(ap, fmt);
	vstring_vsprintf(vp, fmt, ap);
	va_end(ap);
    } else {
	va_start(ap, fmt);
	msg_vprintf(MSG_WARN, fmt, ap);
	va_end(ap);
    }
}

/* v6addr_literal - copy IPv6 literal address from bracketed version */
/*                  Supports both plain addresses and addr/plen's    */

static char *v6addr_literal(const char *pattern)
{
    size_t patlen;
    char *mypattern, *ptr;

    if (pattern == NULL)
	msg_panic("v6_addr_literal: called with NULL pattern pointer");
    if (msg_verbose > 1)
	msg_info("v6addr_literal: input pattern %s", pattern);

    patlen = strlen(pattern);

    /*
     * Note that we allow two different presentation/configuration
     * formats for literal IPv6 (address/prefixlen) combinations.
     * These are [v6addr]/plen and [v6addr/plen]. The second should
     * be avoided and will be deprecated in later Postfix/v6 versions.
     */
    if (*pattern == '[') {
	mypattern = mystrdup(pattern + 1);
	if (pattern[patlen - 1] == ']') {
	    /*
	     * Format: "[v6addr]" or "[v6addr/plen]".
	     */
	    mypattern[patlen - 2] = '\0';
	} else if ((ptr = strchr(mypattern + 1, '/')) != NULL
		   && *--ptr == ']') {
	    /*
	     * Format: "[v6addr]/plen".
	     */
	    while (*ptr)
		ptr++[0] = ptr[1];
	}
    } else {
	mypattern = mystrdup(pattern);
    }

    if (msg_verbose > 1)
	msg_info("v6addr_literal: debracketed to %s", mypattern);

    return (mypattern);
}

/* std_addr_pattern - standardize address pattern */

int std_addr_pattern(int flags, const char *pattern,
		     ADDR_PATTERN **result, VSTRING *warnings)
{
    char   *myname = "std_addr_pattern";
    ADDR_PATTERN *res;
    int     mask;
#ifdef INET6
    int     pf;
    char   *mypattern, *plenp;
    int     bits_per_addr, aierr;
    struct addrinfo hints, *res0;
    struct sockaddr_storage *ss_pattern;

    pf = PF_UNSPEC;
    *result = NULL;

    if (pattern == NULL)
	msg_panic("%s: pattern may not be NULL!", myname);

    /*
     * IPv6 addresses passed as pattern to match_hostaddr should start
     * with a bracket '[' and have a ']' closing. This is as specific
     * as it can get.
     */
    mypattern = v6addr_literal(pattern);
    if (*pattern == '[') {
	pf = PF_INET6;
    } else if (!(flags & (MATCH_FLAG_STRICT_ADDR|MATCH_FLAG_NOLOOKUP))) {
	/*
	 * Return if we find what appears to be a maptype:file entry.
	 * It's up to the caller of this function to handle this.
	 */
	if (strchr(pattern, ':') != NULL) {
	    myfree(mypattern);
	    return (1);
	}
    }
    plenp = split_at(mypattern, '/');

    memset(&hints, 0, sizeof(hints));
    hints.ai_family = pf;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_NUMERICHOST;
    aierr = getaddrinfo(mypattern, NULL, &hints, &res0);
    /*
     * EAI_NONAME happens when the pattern was not supplied in a
     * valid printable form. This is a non-fatal error in strict
     * address pattern maps like the CIDR dictionary.
     */
    if (aierr == EAI_NONAME) {
	if (msg_verbose || (flags & MATCH_FLAG_STRICT_ADDR))
	    warning_msg(warnings,
			"%s: invalid address pattern \"%s\"",
			myname, mypattern);
	myfree(mypattern);
	return (0);
    }
    if (aierr != 0 && aierr != EAI_NONAME)
	msg_fatal("%s: getaddrinfo(%s): %s", myname, mypattern,
		  GAI_STRERROR(aierr));
    pf = res0->ai_family;
    switch (pf) {
    case AF_INET:
	bits_per_addr = BITS_PER_ADDR_V4;
	break;
    case AF_INET6:
	bits_per_addr = BITS_PER_ADDR_V6;
	break;
    default:
	warning_msg(warnings,
		    "%s: unsupported address family %d in lookup result "
		    "of \"%s\"", myname, pf, pattern);
	freeaddrinfo(res0);
	myfree(mypattern);
	return (0);
    }
    ss_pattern = (struct sockaddr_storage *)
		 mymalloc(sizeof(struct sockaddr_storage));
    memcpy(ss_pattern, (const void *)res0->ai_addr, res0->ai_addrlen);
    freeaddrinfo(res0);

    if (plenp != NULL) {
	/*
	 * Split the pattern into an address and a prefix length
	 * We explicitly allow "/0"
	 */
	if (strcmp(plenp, "0")) {
	    mask = atoi(plenp);
	    if (mask <= 0 || mask > bits_per_addr) {
		warning_msg(warnings, "%s: bad net/mask pattern: %s",
			    myname, pattern);
		myfree(mypattern);
		myfree((char *)ss_pattern);
		return (0);
	    }
	} else {
	    mask = 0;
	}
    } else {
	/*
	 * A single address is considered a prefix with maximum prefix length.
	 */
	switch (pf) {
	    case AF_INET:
		mask = BITS_PER_ADDR_V4;
		break;
	    case AF_INET6:
		mask = BITS_PER_ADDR_V6;
		break;
	    default:
		msg_panic("%s: address family %d should not occur here",
			  myname, pf);
	}
    }

    if (flags & MATCH_FLAG_NONNULL_HOST) {
	/*
	 * We require that the host portion of (address/plen) pairs be zero
	 * to reduce the impact of configuration errors.
	 */
	int non_null = 0;

	if (mask != 0 && mask != bits_per_addr) {
	    int bytesl, bits;
	    char *addr = NULL;
	    unsigned char ac;

	    switch (ss_pattern->ss_family) {
	    case AF_INET6:
		addr = (char *)(&((struct sockaddr_in6 *)ss_pattern)->sin6_addr);
		bits_per_addr = BITS_PER_ADDR_V6;
		break;
	    case AF_INET:
		addr = (char *)(&((struct sockaddr_in *)ss_pattern)->sin_addr);
		bits_per_addr = BITS_PER_ADDR_V4;
		break;
	    default:
		msg_panic("%s: address family %d should not occur here",
			  myname, pf);
	    }
	    bytesl = mask / 8;
	    bits = (bits_per_addr - mask) % 8;
	    if (bytesl == bits_per_addr / 8)
		non_null = 1;
	    else
		ac = addr[bytesl];
	    if (bits == 0)
		bits = 8;
	    if (!non_null && ac != (ac & 0xff << bits))
		non_null = 1;
	    while (!non_null && ++bytesl < bits_per_addr / 8)
		non_null = addr[bytesl] != 0;
	}
	if (non_null) {
	    warning_msg(warnings,
			"%s: net/mask pattern \"%s/%s\" "
			"with non-null host pattern",
			myname, mypattern, plenp);
	    myfree(mypattern);
	    myfree((char *)ss_pattern);
	    return (0);
	}
    }

#else /* INET6 */

    char *mypattern, *plenp;
    int bits;
    unsigned long addr, addr0;
    struct sockaddr_in *ss_pattern;

    *result = NULL;

    if (!(flags & MATCH_FLAG_STRICT_ADDR) && strchr(pattern, ':') != 0)
	return (1);

    mypattern = mystrdup(pattern);
    plenp = split_at(mypattern, '/');
    if (plenp == NULL) {
	bits = BITS_PER_ADDR_V4;
    } else {
	bits = atoi(plenp);
	if (bits <= 0 || bits > BITS_PER_ADDR_V4)
	warning_msg(warnings,
		    "%s: bad net/mask pattern: %s",
		    myname, pattern);
	myfree(mypattern);
	return (0);
    }

    addr = inet_addr(mypattern);
    addr0 = htonl(0xffffffff << (BITS_PER_ADDR_V4 - bits));
    if ((flags & MATCH_FLAG_NONNULL_HOST) && (addr & ~addr0)) {
	warning_msg(warnings,
		    "%s: net/mask pattern \"%s/%s\" with "
		    "non-null host portion",
		    myname, mypattern, plenp);
	myfree(mypattern);
	return (0);
    }

    /*
     * We make a sockaddr_in, but we don't use any of the fields
     * except the sin_addr member. Sockaddrs are used to create
     * an API that's closer to AF-independence.
     */
    ss_pattern = (struct sockaddr_in *)mymalloc(sizeof(struct sockaddr_in));
    memset(ss_pattern, 0, sizeof(*ss_pattern));
    ss_pattern->sin_family = AF_INET;
    ss_pattern->sin_addr.s_addr = addr;

#endif	/* INET6 */

    res = addr_pattern_init();
    res->addr = (struct sockaddr *)ss_pattern;
    res->masklen = mask;
    res->opattern = mystrdup(pattern);
    res->pattern = mypattern;
    *result = res;

    return (1);
}

/* match_hostaddr - match host by address */

/* XXX: the IPv4-only version does not yet use std_addr_pattern --dean */

int     match_hostaddr(int unused_flags, const char *addr, const char *pattern)
{
    char   *myname = "match_hostaddr";
#ifdef INET6
    size_t  patlen;
    char   *plenp;
    int     aierr, res, ret, mask;
    struct addrinfo hints, *res0;
    struct sockaddr_storage ss_addr, ss_mask;
    ADDR_PATTERN *mask_info;

    ret = 0;
    if (msg_verbose)
	msg_info("%s: %s ~? %s", myname, addr, pattern);

    memset(&hints, 0, sizeof(hints));
    hints.ai_family = PF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_NUMERICHOST;
    aierr = getaddrinfo(addr, NULL, &hints, &res0);
    /*
     * The access maps checks run both hostname and address through this.
     * E.g. the CIDR map checks both the hostname and address. Checking the
     * hostname in a CIDR map will yield no result but may not give an
     * EAI_NONAME error since it is correct that the hostname cannot be
     * interpreted numerically.
     */
    if (aierr != 0 && aierr != EAI_NONAME)
	msg_fatal("%s: getaddrinfo(%s): %s", myname, addr, GAI_STRERROR(aierr));
    memcpy(&ss_addr, (const void *)res0->ai_addr, res0->ai_addrlen);
    freeaddrinfo(res0);

    res = std_addr_pattern(MATCH_FLAG_NONE, pattern, &mask_info, NULL);
    if (mask_info == NULL) {
	/*
	 * Try dictionary lookup. This can be case insensitive.
	 */
	if (res && strchr(pattern, ':') != 0) {
	    if (dict_lookup(pattern, addr) != NULL)
		return 1;
	}
	return 0;
    }
    
    /*
     * Try an exact match with the host address (IPv4 only)
     */
    if (mask_info->addr->sa_family == AF_INET &&
			    strcasecmp(pattern, addr) == 0) {
	addr_pattern_free(mask_info);
	return 1;
    }

    res = match_sockaddr((struct sockaddr *)&ss_addr,
			 mask_info->addr, mask_info->masklen);
    addr_pattern_free(mask_info);
    return (res != 0);
    

#else
    unsigned int mask_shift;
    unsigned long mask_bits;
    unsigned long net_bits;
    unsigned long addr_bits;
    struct in_addr net_addr;

    if (msg_verbose)
	msg_info("%s: %s ~? %s", myname, addr, pattern);

    if (addr[strspn(addr, "01234567890./:")] != 0)
	return (0);

    /*
     * Try dictionary lookup. This can be case insensitive. XXX Probably
     * should also try again after stripping least significant octets.
     */
    if (strchr(pattern, ':') != 0)
    {
	if (dict_lookup(pattern, addr) != 0)
	    return (1);
	if (dict_errno != 0)
	    msg_fatal("%s: table lookup problem", pattern);
	return (0);
    }

    /*
     * Try an exact match with the host address.
     */
    if (strcasecmp(addr, pattern) == 0) {
	return (1);
    }

    /*
     * In a net/mask pattern, the mask is specified as the number of bits of
     * the network part.
     */
	    
    if (match_parse_mask(pattern, &net_bits, &mask_shift)) {
	addr_bits = inet_addr(addr);
	if (addr_bits == INADDR_NONE)
	    msg_fatal("%s: bad address argument: %s", myname, addr);
	mask_bits = mask_shift > 0 ?
		htonl((0xffffffff) << (BITS_PER_ADDR_V4 - mask_shift)) : 0;
	if ((addr_bits & mask_bits) == net_bits)
	    return 1;
	if (net_bits & ~mask_bits) {
	    net_addr.s_addr = (net_bits & mask_bits);
	    msg_fatal("net/mask pattern %s has a non-null host portion; "
		      "specify %s/%d if this is really what you want",
		      pattern, inet_ntoa(net_addr), mask_shift);
	}
    }
    return (0);
#endif
}

int
match_sockaddr(const struct sockaddr *addr, const struct sockaddr *mask,
	       int masklen)
{
    /*
     * I generally hate to do so, but this function just asks for
     * #ifdef INET6... address comparison in the IPv4 only case is
     * utterly trivial, completely unlike the mixed AF case.
     */
#ifdef INET6
    if (addr->sa_family == AF_INET) {
	if (mask->sa_family == AF_INET6) {
	    if (IN6_IS_ADDR_V4MAPPED(
			&((struct sockaddr_in6 *)mask)->sin6_addr)) {
	        /* IPv4 address but IPv4-mapped-IPv6 netmask... */
		if (masklen < 0 || masklen > BITS_PER_ADDR_V4)
		    return 0;
	        return mask_comp(&((struct sockaddr_in *)addr)->sin_addr.s_addr,
		    &((struct sockaddr_in6 *)mask)->sin6_addr.s6_addr[12],
		    masklen);
	    }
	    /* IPv4 address yet IPv6 mask. No match */
	    return 0;
        }
	/* IPv4 address, IPv4 netmask */
	if (masklen < 0 || masklen > BITS_PER_ADDR_V4)
	    return 0;
	return mask_comp(&((struct sockaddr_in *)addr)->sin_addr.s_addr,
			 &((struct sockaddr_in *)mask)->sin_addr.s_addr,
			 masklen);
    } else if (addr->sa_family == AF_INET6) {
	/* IPv6 address, IPv6 netmask */
	struct sockaddr_in6 *addr6, *mask6;
	addr6 = (struct sockaddr_in6 *)addr;
	mask6 = (struct sockaddr_in6 *)mask;

	if (IN6_IS_ADDR_V4MAPPED(&addr6->sin6_addr)) {
	    /* V4-mapped IPv6 address */
	    struct sockaddr_in addr4;
	    memset(&addr4, 0, sizeof(addr4));
#ifdef HAS_SA_LEN
	    addr4.sin_len = sizeof(addr4);
#endif
	    addr4.sin_family = AF_INET;
	    memcpy(&addr4.sin_addr.s_addr, &addr6->sin6_addr.s6_addr[12], 4);
	    if (masklen > BITS_PER_ADDR_V4 && masklen <= BITS_PER_ADDR_V6)
		masklen -= BITS_PER_ADDR_V6 - BITS_PER_ADDR_V4;
	    return match_sockaddr((struct sockaddr *)&addr4, mask, masklen);
	}
	/* True IPv6, finally... */
        if (masklen < 0 || masklen > BITS_PER_ADDR_V6)
	    return 0;
	if (mask->sa_family != AF_INET6 ||
		IN6_IS_ADDR_V4MAPPED(&mask6->sin6_addr))
	    return 0;
#ifdef INET6_KAME
	if (IN6_IS_ADDR_SITELOCAL(&addr6->sin6_addr))
	    if (!IN6_IS_ADDR_SITELOCAL(&mask6->sin6_addr) ||
		    addr6->sin6_scope_id != mask6->sin6_scope_id)
		return 0;
	if (IN6_IS_ADDR_LINKLOCAL(&addr6->sin6_addr))
	    if (!IN6_IS_ADDR_LINKLOCAL(&mask6->sin6_addr) ||
		    addr6->sin6_scope_id != mask6->sin6_scope_id)
		return 0;
#endif
	return mask_comp(&addr6->sin6_addr.s6_addr,
			 &mask6->sin6_addr.s6_addr,
			 masklen);
    }
    /* Unsupported address family */
    return 0;
#else /* INET6 */
    /*
     * Trivial for IPv4...
     */
    return (addr->sa_family == mask->sa_family &&
	    ((struct sockaddr_in *)addr)->sin_addr.s_addr ==
	    ((struct sockaddr_in *)mask)->sin_addr.s_addr);
#endif /* INET6 */
}

static int
mask_comp(void *addr, void *mask, int masklen)
{
    int bytes, bit;

    bytes = masklen / 8;
    bit = 8 - masklen % 8;
    if (memcmp(addr, mask, bytes) != 0)
	return 0;
    if (bit != 8) {
	char *a = addr, *b = mask;
	if ((a[bytes] & (0xff << bit)) != (b[bytes] & (0xff << bit)))
	    return 0;
    }
    return 1;
}

ADDR_PATTERN *
addr_pattern_init() {
    ADDR_PATTERN *p;
    p = (ADDR_PATTERN *)mymalloc(sizeof(ADDR_PATTERN));
    memset(p, 0, sizeof(ADDR_PATTERN));
    return p;
}

void
addr_pattern_free(ADDR_PATTERN *p) {
    if (p->addr) myfree((char *)p->addr);
    if (p->pattern) myfree(p->pattern);
    if (p->opattern) myfree(p->opattern);
    myfree((char *)p);
}

